/*
 * The MIT License (MIT)
 *
 * Copyright (c) 2013-2023 Chaosdorf e.V.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package de.chaosdorf.meteroid.di

import androidx.room.withTransaction
import de.chaosdorf.mete.v1.MeteApiV1Factory
import de.chaosdorf.meteroid.MeteroidDatabase
import de.chaosdorf.meteroid.RepositorySyncHandler
import de.chaosdorf.meteroid.SyncHandler
import de.chaosdorf.meteroid.model.Drink
import de.chaosdorf.meteroid.model.Server
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.flow.StateFlow
import kotlinx.coroutines.launch


interface MainLayoutViewModelFactory {
  fun newInstance(
    scope: CoroutineScope, db: MeteroidDatabase, server: Server, onOpenServerSelection: () -> Unit
  ): MainLayoutViewModel = MainLayoutViewModelImpl(scope, db, server, onOpenServerSelection)
}

interface MainLayoutViewModel {
  val server: Server
  val syncState: StateFlow<SyncHandler.State>
  val drinkListViewModel: DrinkListViewModel

  fun openServerSelection()
}

class MainLayoutViewModelImpl(
  scope: CoroutineScope,
  db: MeteroidDatabase,
  override val server: Server,
  private val onOpenServerSelection: () -> Unit
) : MainLayoutViewModel {
  private val api = MeteApiV1Factory.newInstance(server.url)
  override val drinkListViewModel = DrinkListViewModelImpl(scope, db.drinks())
  private val syncHandler = RepositorySyncHandler(db::withTransaction, db.drinks()) {
    api.listDrinks().map { Drink.fromModelV1(it) }
  }
  override val syncState: StateFlow<SyncHandler.State> = syncHandler.state

  init {
    scope.launch {
      syncHandler.doSync()
    }
  }

  override fun openServerSelection() {
    onOpenServerSelection()
  }
}
