/*
 * The MIT License (MIT)
 *
 * Copyright (c) 2013-2023 Chaosdorf e.V.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package de.chaosdorf.meteroid.di

import android.content.Context
import androidx.room.Room
import de.chaosdorf.mete.v1.MeteApiV1Factory
import de.chaosdorf.meteroid.MeteroidDatabase
import de.chaosdorf.meteroid.Repository
import de.chaosdorf.meteroid.SyncHandler
import de.chaosdorf.meteroid.model.Server
import de.chaosdorf.meteroid.model.ServerId
import de.chaosdorf.meteroid.util.findBestIcon
import de.chaosdorf.meteroid.util.resolve
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.flow.MutableStateFlow
import kotlinx.coroutines.flow.SharingStarted
import kotlinx.coroutines.flow.StateFlow
import kotlinx.coroutines.flow.combine
import kotlinx.coroutines.flow.flatMapLatest
import kotlinx.coroutines.flow.flowOf
import kotlinx.coroutines.flow.stateIn
import kotlinx.coroutines.launch

interface RootViewModelFactory {
  fun newInstance(
    scope: CoroutineScope, context: Context
  ): RootViewModel = RootViewModelImpl(scope, context)
}

interface RootViewModel {
  val route: StateFlow<RootRoute>
}

class ServerListSyncHandler(
  private val repository: Repository<ServerId, Server>,
) : SyncHandler {
  private val _state = MutableStateFlow<SyncHandler.State>(SyncHandler.State.Idle)
  override val state = _state

  override suspend fun doSync() {
    _state.value = SyncHandler.State.Syncing
    for (server in repository.getAll()) {
      val api = MeteApiV1Factory.newInstance(server.url)
      val manifest = api.getManifest()
      val updated = server.copy(
        name = manifest?.name,
        logoUrl = manifest?.findBestIcon()?.resolve(server.url)
      )
      repository.save(updated)
    }
    _state.value = SyncHandler.State.Idle
  }
}

class RootViewModelImpl(
  scope: CoroutineScope, context: Context
) : RootViewModel {
  private val db = Room.databaseBuilder(context, MeteroidDatabase::class.java, "mete").build()

  private val syncHandler = ServerListSyncHandler(db.server())

  private val setupViewModelFactory: SetupViewModelFactory = object : SetupViewModelFactory {}

  private val mainLayoutViewModelFactory: MainLayoutViewModelFactory =
    object : MainLayoutViewModelFactory {}

  private val _serverSelectionOpen = MutableStateFlow(false)
  private val _serverId: MutableStateFlow<ServerId?> = MutableStateFlow(null)
  private val _server: StateFlow<Server?> = _serverId.flatMapLatest {
    it?.let { db.server().getFlow(it) } ?: flowOf(null)
  }.stateIn(scope, SharingStarted.WhileSubscribed(), null)
  override val route: StateFlow<RootRoute> =
    combine(_server, _serverSelectionOpen) { server, serverSelectionOpen ->
      if (server == null || serverSelectionOpen) {
        RootRoute.Setup(
          setupViewModelFactory.newInstance(
            scope,
            db,
            db.server(),
            server != null,
            ::onSelect,
            ::onCloseServerSelection
          )
        )
      } else {
        RootRoute.MainLayout(
          mainLayoutViewModelFactory.newInstance(
            scope,
            db,
            server,
            ::onOpenServerSelection
          )
        )
      }
    }.stateIn(scope, SharingStarted.WhileSubscribed(), RootRoute.Init)

  init {
    scope.launch { syncHandler.doSync() }
  }

  private fun onSelect(server: ServerId) {
    _serverId.value = server
    _serverSelectionOpen.value = false
  }

  private fun onOpenServerSelection() {
    _serverSelectionOpen.value = true
  }

  private fun onCloseServerSelection() {
    _serverSelectionOpen.value = true
  }
}

sealed class RootRoute {
  data object Init : RootRoute()
  data class Setup(val viewModel: SetupViewModel) : RootRoute()
  data class MainLayout(val viewModel: MainLayoutViewModel) : RootRoute()
}
