/*
 * The MIT License (MIT)
 *
 * Copyright (c) 2013-2023 Chaosdorf e.V.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package de.chaosdorf.meteroid.di

import de.chaosdorf.meteroid.Repository
import de.chaosdorf.meteroid.model.Server
import de.chaosdorf.meteroid.model.ServerId
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.flow.SharingStarted
import kotlinx.coroutines.flow.StateFlow
import kotlinx.coroutines.flow.stateIn
import kotlinx.coroutines.launch


interface ServerSelectionViewModelFactory {
  fun newInstance(
    scope: CoroutineScope,
    repository: Repository<ServerId, Server>,
    hasSelectedServer: Boolean,
    onAddServer: () -> Unit,
    onSelect: (server: ServerId) -> Unit,
    onClose: () -> Unit
  ) = ServerSelectionViewModelImpl(
    scope,
    repository,
    hasSelectedServer,
    onAddServer,
    onSelect,
    onClose
  )
}

interface ServerSelectionViewModel {
  val servers: StateFlow<List<Server>>
  val hasSelectedServer: Boolean
  fun addServer()
  fun select(server: ServerId)
  fun remove(server: ServerId)
  fun close()
}

class ServerSelectionViewModelImpl(
  private val scope: CoroutineScope,
  private val repository: Repository<ServerId, Server>,
  override val hasSelectedServer: Boolean,
  private val onAddServer: () -> Unit,
  private val onSelect: (server: ServerId) -> Unit,
  private val onClose: () -> Unit
) : ServerSelectionViewModel {
  override val servers: StateFlow<List<Server>> =
    repository.getAllFlow().stateIn(scope, SharingStarted.WhileSubscribed(), emptyList())

  override fun addServer() {
    onAddServer()
  }

  override fun select(server: ServerId) {
    onSelect(server)
  }

  override fun remove(server: ServerId) {
    scope.launch {
      repository.delete(server)
    }
  }

  override fun close() {
    onClose()
  }
}
