/*
 * The MIT License (MIT)
 *
 * Copyright (c) 2013-2023 Chaosdorf e.V.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package de.chaosdorf.meteroid.di

import androidx.room.withTransaction
import de.chaosdorf.mete.PwaManifest
import de.chaosdorf.meteroid.MeteroidDatabase
import de.chaosdorf.meteroid.Repository
import de.chaosdorf.meteroid.model.Server
import de.chaosdorf.meteroid.model.ServerId
import de.chaosdorf.meteroid.util.findBestIcon
import de.chaosdorf.meteroid.util.resolve
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.flow.MutableStateFlow
import kotlinx.coroutines.flow.SharingStarted
import kotlinx.coroutines.flow.StateFlow
import kotlinx.coroutines.flow.combine
import kotlinx.coroutines.flow.map
import kotlinx.coroutines.flow.stateIn
import kotlinx.coroutines.launch


interface SetupViewModelFactory {
  fun newInstance(
    scope: CoroutineScope,
    db: MeteroidDatabase,
    repository: Repository<ServerId, Server>,
    hasSelectedServer: Boolean,
    onSelect: (server: ServerId) -> Unit,
    onClose: () -> Unit
  ): SetupViewModel = SetupViewModelImpl(
    scope, db, repository, hasSelectedServer, onSelect, onClose
  )
}

interface SetupViewModel {
  val route: StateFlow<SetupRoute>
}

class SetupViewModelImpl(
  private val scope: CoroutineScope,
  private val db: MeteroidDatabase,
  private val repository: Repository<ServerId, Server>,
  private val hasSelectedServer: Boolean,
  private val onSelect: (server: ServerId) -> Unit,
  private val onClose: () -> Unit
) : SetupViewModel {
  private val addServerViewModelFactory = object : AddServerViewModelFactory {}
  private val serverSelectionViewModelFactory = object : ServerSelectionViewModelFactory {}
  private val isFirstServer = repository.getAllFlow().map(List<Server>::isEmpty)
  private val isAddingServer = MutableStateFlow(false)
  override val route: StateFlow<SetupRoute> =
    combine(isFirstServer, isAddingServer) { isFirstServer, addingServer ->
      if (addingServer || isFirstServer) {
        SetupRoute.AddServer(
          addServerViewModelFactory.newInstance(
            scope, isFirstServer, ::onSubmit, ::onCancel
          )
        )
      } else {
        SetupRoute.ServerSelection(
          serverSelectionViewModelFactory.newInstance(
            scope, repository, hasSelectedServer, ::onAddServer, onSelect, onClose
          )
        )
      }
    }.stateIn(
      scope, SharingStarted.WhileSubscribed(), SetupRoute.ServerSelection(
        serverSelectionViewModelFactory.newInstance(
          scope, repository, hasSelectedServer, ::onAddServer, onSelect, onClose
        )
      )
    )

  private fun onAddServer() {
    isAddingServer.value = true
  }

  private fun onSubmit(url: String, manifest: PwaManifest?) {
    isAddingServer.value = false
    scope.launch {
      db.withTransaction {
        val lastId = repository.getAll().map(Server::id).maxByOrNull(ServerId::value)?.value ?: 0
        repository.save(
          Server(
            id = ServerId(lastId + 1),
            url = url,
            name = manifest?.name,
            logoUrl = manifest?.findBestIcon()?.resolve(url)
          )
        )
      }
    }
  }

  private fun onCancel() {
    isAddingServer.value = false
  }
}

sealed class SetupRoute {
  data class ServerSelection(val viewModel: ServerSelectionViewModel) : SetupRoute()
  data class AddServer(val viewModel: AddServerViewModel) : SetupRoute()
}
