/*
 * The MIT License (MIT)
 *
 * Copyright (c) 2013-2023 Chaosdorf e.V.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package de.chaosdorf.meteroid.model

import androidx.room.Dao
import androidx.room.Entity
import androidx.room.Insert
import androidx.room.OnConflictStrategy
import androidx.room.PrimaryKey
import androidx.room.Query
import de.chaosdorf.mete.DrinkId
import de.chaosdorf.mete.v1.DrinkModelV1
import de.chaosdorf.meteroid.Repository
import kotlinx.coroutines.flow.Flow
import kotlinx.datetime.Instant

@Entity
data class Drink(
  @PrimaryKey
  val id: DrinkId,
  val name: String,
  val volume: Double,
  val caffeine: Int?,
  val price: Double,
  val active: Boolean,
  val createdAt: Instant,
  val updatedAt: Instant,
  val logoUrl: String,
  val logoFileName: String,
  val logoContentType: String,
  val logoFileSize: Long,
  val logoUpdatedAt: Instant
) {
  companion object {
    fun fromModelV1(value: DrinkModelV1) = Drink(
      value.id,
      value.name,
      value.bottleSize,
      value.caffeine,
      value.price,
      value.active,
      value.createdAt,
      value.updatedAt,
      value.logoUrl,
      value.logoFileName,
      value.logoContentType,
      value.logoFileSize,
      value.logoUpdatedAt
    )
  }
}

@Dao
interface DrinkDao : Repository<DrinkId, Drink> {
  override fun getKey(value: Drink): DrinkId = value.id

  @Query("SELECT * FROM Drink WHERE id = :id LIMIT 1")
  override suspend fun get(id: DrinkId): Drink?

  @Query("SELECT * FROM Drink WHERE id = :id LIMIT 1")
  override fun getFlow(id: DrinkId): Flow<Drink?>

  @Query("SELECT * FROM Drink")
  override suspend fun getAll(): List<Drink>

  @Query("SELECT * FROM Drink")
  override fun getAllFlow(): Flow<List<Drink>>

  @Insert(onConflict = OnConflictStrategy.REPLACE)
  override suspend fun save(drink: Drink)

  @Query("DELETE FROM Drink WHERE id = :id")
  override suspend fun delete(id: DrinkId)

  @Query("DELETE FROM Drink")
  override suspend fun deleteAll()
}
